import json
import os
import logging
from distutils.version import StrictVersion
from shutil import copyfile
from urllib import request

from carball.rattletrap.rattletrap_utils import get_rattletrap_path, get_binary_version, \
    get_all_binaries, get_highest_binary, get_rattletrap_binaries, copy_cloud_over_to_rattletrap

log = logging.getLogger(__name__)


def update_rattletrap():
    path = get_rattletrap_path()

    cur_ver = StrictVersion('0.0.0')
    binaries = get_all_binaries(path)

    try:
        response = request.urlopen('https://api.github.com/repos/tfausak/rattletrap/releases/latest')
        js = json.loads(response.read())
        github_ver = StrictVersion(js['name'])
    except:
        log.warning('Unable to download rattletrap copying backup')
        # unable to download a new rattletrap version so we should just copy our own
        github_ver = StrictVersion(cur_ver)
        copy_cloud_over_to_rattletrap(binaries)
        return

    if len(binaries) > 0:
        binary = get_highest_binary(binaries)
        if 'cloud' in binary:
            log.warning('Cloud parser is highest binary')
            copy_cloud_over_to_rattletrap(binaries)
        cur_ver = get_binary_version(binary)
    update = github_ver > cur_ver
    log.info(f'GitHub version: {js["name"]}\n'
             f'Current version: {cur_ver}\n'
             f'Update? {update}')
    if update:
        for file in get_rattletrap_binaries(path):
            os.remove(os.path.join(path, file))

        for asset in js['assets']:
            log.info('Downloading {}'.format(asset['name']))
            new_file = os.path.join(path, asset['name'])
            request.urlretrieve(asset['browser_download_url'], filename=new_file)
            log.info('making file executable', new_file)
            os.chmod(new_file, 0o777)


if __name__ == "__main__":
    update_rattletrap()



import inspect
import os
import fnmatch
import logging
from distutils.version import StrictVersion
from shutil import copyfile
from typing import List, Union

log = logging.getLogger(__name__)


def get_rattletrap_path() -> Union[bytes, str]:
    filename = inspect.getframeinfo(inspect.currentframe()).filename
    path = os.path.dirname(os.path.abspath(filename))
    return path


def get_rattletrap_binaries(path: Union[bytes, str]) -> List[Union[bytes, str]]:
    files = os.listdir(path)
    binaries = [f for f in files if not f.endswith('.py') and (fnmatch.fnmatch(f, "*rattletrap*"))]
    log.debug(f'Discovered rattletrap binaries: {binaries}')
    return binaries


def get_all_binaries(path: Union[bytes, str]) -> List[Union[bytes, str]]:
    files = os.listdir(path)
    binaries = [f for f in files if not f.endswith('.py') and (fnmatch.fnmatch(f, "*rattletrap*") or
                                                               fnmatch.fnmatch(f, "*cloud_parser*"))]
    log.debug(f'Discovered rattletrap binaries: {binaries}')
    return binaries


def download_rattletrap():
    from carball.rattletrap.check_rattletrap_version import update_rattletrap
    update_rattletrap()


def get_binary_version(filename: str) -> StrictVersion:
    try:
        return StrictVersion(filename.split('-')[1])
    except:
        return StrictVersion('0.0.0')


def get_highest_binary(binaries):
    if len(binaries) == 1:
        return binaries[0]
    binaries.sort(key=lambda value: get_binary_version(str(value)))
    binary = binaries[-1]
    log.debug(binary)
    return binary


def get_binary_for_platform(platform, binaries):
    if platform == 'Windows':
        binaries = [f for f in binaries if f.endswith('.exe')]
    elif platform == 'Linux':
        binaries = [f for f in binaries if 'linux' in f]
    elif platform == 'Darwin':
        binaries = [f for f in binaries if 'osx' in f]
    else:
        raise Exception('Unknown platform, unable to process replay file.')

    return get_highest_binary(binaries)


def copy_cloud_over_to_rattletrap(binaries):
    cloud_binary = list(filter(lambda file_name: 'cloud_parser' in str(file_name), binaries))[0]
    version_info = get_binary_version(cloud_binary)
    linux_name = 'rattletrap-' + str(version_info) + '-linux'
    copyfile(os.path.join(get_rattletrap_path(), cloud_binary),
             os.path.join(get_rattletrap_path(), linux_name))
    os.chmod(os.path.join(get_rattletrap_path(), linux_name), 0o777)
    log.info("copied " + cloud_binary + " to " + linux_name)



from .actor import *

REPLICATED_RB_STATE_KEY = 'TAGame.RBActor_TA:ReplicatedRBState'

_HANDLERS = [
    GameInfoHandler,
    GameEventHandler,
    BallHandler,
    PlayerHandler,
    TeamHandler,
    CarHandler,
    JumpHandler,
    DodgeHandler,
    DoubleJumpHandler,
    BoostHandler,
    BoostPickupHandler,
    CameraSettingsHandler,
    RumbleItemHandler,
    PlatformHandler
]

_PRIORITY_HANDLERS = [
    GameInfoHandler,
    PlayerHandler,
    CarHandler
]

# These handlers will also handle frames with a delta of 0 to match the old implementation
# TODO check if this is needed
_0_DELTA_HANDLERS = [
    GameInfoHandler,
    PlayerHandler
]


def parse_frames(game):
    """
    :return: all_data = {
        'player_ball_data': player_ball_data,
        'player_dicts': player_dicts,
        'team_dicts': team_dicts,
        'frames_data': frames_data,
        'cameras_data': cameras_data
        'demos_data': demos_data
    }

    player_ball_data format:
    {
    'ball': {frame_number: {pos_x, pos_y ...}, f_no2: {...} ...,}
    player_actor_id: {
        frame_number: {
            pos_x, pos_y ...,
            throttle, steer, ...,
            ping, ball_cam
        },
        f_no2: {...} ...,
    }

    currently implemented:
        inputs: posx, posy, posz, rotx, roty, rotz, vx, vy, vz, angvx, angy, angvz, boost_amt
        outputs: throttle, steer, handbrake, boost, jump, doublejump, dodge

    player_dicts  = {player_actor_id : {actor_data}, player_actor_id_2: {actor_data_2}}
    team_dicts = {team_actor_id: {actor_data, 'colour':'blue'/'orange', also includes name}
    frames_data = {frame_number: {time, delta, seconds_remaining, is_overtime, ball_has_been_hit}
    cameras_data = {player_actor_id: actor_data}
    demos_data = {frame_number: demolish_data}

    """
    parser = FrameParser(game.replay_data, game)
    parser.parse_frames()

    player_ball_data = parser.player_data
    player_ball_data['ball'] = parser.ball_data

    return {
        'player_ball_data': player_ball_data,
        'player_dicts': parser.player_dicts,
        'car_dicts': parser.car_dicts,
        'team_dicts': parser.team_dicts,
        'frames_data': parser.frames_data,
        'cameras_data': parser.cameras_data,
        'demos_data': parser.demos_data,
        'game_info_actor': parser.game_info_actor,
        'soccar_game_event_actor': parser.soccar_game_event_actor,
        'parties': parser.parties,
        'dropshot': parser.dropshot
    }


class FrameParser(object):

    def __init__(self, replay_frames, game):
        self.replay_frames = replay_frames
        self.game = game
        self.replay_version = game.replay_version

        self.game_info_actor = None
        self.soccar_game_event_actor = None

        self.frames_data = {}
        self.ball_data = {}
        self.player_data = {}

        # dictionaries to contain data in frames
        self.parties = {}
        self.player_dicts = {}
        self.team_dicts = {}
        self.car_dicts = {}

        self.player_car_ids = {}  # player_actor_id: car_actor_id
        self.car_player_ids = {}  # car_actor_id: player_actor_id

        self.cameras_data = {}  # player_actor_id: actor_data
        self.demos_data = []  # frame_number: demolish_data

        # stores car_actor_ids to collect data for at each frame
        self.current_car_ids_to_collect = []

        self.actors = {}

        self.dropshot = {
            'tile_states': {},
            'damage_events': {},
            'ball_state': 0,
            'ball_events': [],
            'tile_frames': {}
        }

    def parse_frames(self):

        self.actors = {}
        handlers = [dict() for _ in range(len(_PRIORITY_HANDLERS) + 1)]
        handled_actors = set()

        destroyed_actors = set()

        current_goal_number = 0

        for i, frame in enumerate(self.replay_frames):
            time = frame['time']
            delta = frame['delta']

            destroyed_actors.clear()

            for replication in frame['replications']:
                actor_id = replication['actor_id']['value']
                actor_status = next(iter(replication['value'].keys()))

                if actor_status == 'spawned':
                    self.actors[actor_id] = {
                        'Id': actor_id,
                        'TypeName': replication['value']['spawned']['object_name'],
                        'ClassName': replication['value']['spawned']['class_name'],
                        'Name': replication['value']['spawned'].get('name', None),
                    }

                    handler = next(filter(lambda handler_cls: handler_cls.can_handle(self.actors[actor_id]),
                                          _HANDLERS), None)

                    if handler is not None:
                        try:
                            priority = _PRIORITY_HANDLERS.index(handler)
                        except ValueError:
                            priority = len(_PRIORITY_HANDLERS)

                        handlers[priority][actor_id] = handler(self), handler in _0_DELTA_HANDLERS
                        handled_actors.add(actor_id)

                elif actor_status == 'updated':
                    if actor_id not in self.actors:
                        continue

                    actor = self.actors[actor_id]
                    # update properties
                    for prop in replication['value']['updated']:
                        actor[prop['name']] = find_actual_value(prop['value'])

                elif actor_status == 'destroyed':
                    if actor_id in handled_actors:
                        destroyed_actors.add(actor_id)

            # apply destroy handlers
            for actor_id in destroyed_actors:
                for handler_group in handlers:
                    handler_group.pop(actor_id, None)
                handled_actors.remove(actor_id)
                self.player_car_ids.pop(actor_id, None)
                self.car_player_ids.pop(actor_id, None)
                self.actors.pop(actor_id, None)

            # stop data collection after goal
            try:
                if i > self.game.goals[current_goal_number].frame_number:
                    # set all players to sleeping after goal
                    for car_actor_id in self.car_player_ids:
                        try:
                            car = self.actors[car_actor_id]
                            car[REPLICATED_RB_STATE_KEY]['Sleeping'] = True
                        except KeyError as e:
                            # Ignore the case where the car does not have a REPLICATED_RB_STATE_KEY
                            pass
                    current_goal_number += 1
            except IndexError:
                # after last goal.
                pass

            # apply the update handlers
            for handler_group in handlers:
                for actor_id, handler_tuple in handler_group.items():
                    handler = handler_tuple[0]

                    # skip 0 delta frames, except for these two handlers (matches old implementation)
                    if handler_tuple[1] or delta != 0:
                        handler.update(self.actors[actor_id], i, time, delta)

            self.current_car_ids_to_collect.clear()


_TYPES = ['int', 'boolean', 'string', 'byte', 'str', 'name']


def find_actual_value(value_dict: dict) -> dict or int or bool or str:
    if value_dict is None:
        return None
    if 'value' in value_dict:
        value_dict = value_dict['value']
    key, value = next(iter(value_dict.items()))
    if key == 'flagged_int':
        return value['int']
    if key in _TYPES:
        return value
    return value_dict



import numpy as np
import pandas as pd
from carball.generated.api import game_pb2

# The first two values are X,Y. The last value is (RLBot_label +1) and multiplied by 10 if its a bigboost.
# Bigboosts are *10 so that all big boost values are larger than all small boost values, for easy querying.
# If nobody needs these to be RLBot labels, this can be simplified.
BIG_BOOST_POSITIONS = np.array([
    (3072, -4096, 50),
    (-3072, -4096, 40),
    (3584, 0, 190),
    (-3584, 0, 160),
    (3072, 4096, 300),
    (-3072, 4096, 310)])

SMALL_BOOST_POSITIONS = np.array([
    (0.0, -4240.0, 1),
    (-1792.0, -4184.0, 2),
    (1792.0, -4184.0, 3),
    (- 940.0, -3308.0, 6),
    (940.0, -3308.0, 7),
    (0.0, -2816.0, 8),
    (-3584.0, -2484.0, 9),
    (3584.0, -2484.0, 10),
    (-1788.0, -2300.0, 11),
    (1788.0, -2300.0, 12),
    (-2048.0, -1036.0, 13),
    (0.0, -1024.0, 14),
    (2048.0, -1036.0, 15),
    (-1024.0, 0.0, 17),
    (1024.0, 0.0, 18),
    (-2048.0, 1036.0, 20),
    (0.0, 1024.0, 21),
    (2048.0, 1036.0, 22),
    (-1788.0, 2300.0, 23),
    (1788.0, 2300.0, 24),
    (-3584.0, 2484.0, 25),
    (3584.0, 2484.0, 26),
    (0.0, 2816.0, 27),
    (- 940.0, 3310.0, 28),
    (940.0, 3308.0, 29),
    (-1792.0, 4184.0, 32),
    (1792.0, 4184.0, 33),
    (0.0, 4240.0, 34)
])

BIG_BOOST_RADIUS = 208
SMALL_BOOST_RADIUS = 149  # 144 doesn't work for some pickups that are very close to the edge.
BIG_BOOST_HEIGHT = 168
SMALL_BOOST_HEIGHT = 165
# Choosing how many frames to be open to setting a pickup. Back is for when the player is ahead of the server (usually smaller)
LAG_BACK = 6
LAG_FORWARD = 14

BOOST_POSITIONS = np.concatenate((BIG_BOOST_POSITIONS, SMALL_BOOST_POSITIONS))


class PickupAnalysis:

    @staticmethod
    def add_pickups(proto_game: game_pb2.Game, data_frame: pd.DataFrame):

        for player in proto_game.players:
            player_vals_df = data_frame[player.name][['pos_x', 'pos_y', 'pos_z', 'boost']].copy()
            player_vals_df['boost'] /= 2.55
            player_vals_df['boost'] = player_vals_df['boost'].round(5)
            player_vals_df = player_vals_df.dropna(axis=0, how='all')
            player_vals_df = player_vals_df.fillna(0)
            player_vals_df['boost_collect'] = get_boost_collect(player_vals_df)
            data_frame[player.name, 'boost_collect'] = player_vals_df['boost_collect']
        return

    @staticmethod
    def something(proto_game: game_pb2.Game, data_frame: pd.DataFrame):
        pass


def get_boost_collect(player_vals_df):
    # Get a series with indexes as a subset of the indexes of df, values being pad label picked up.
    # Iterate through every pad, label each frame in the path with which boost pad it was in range of.
    df = player_vals_df.copy()
    path = df.drop(['pos_z', 'boost'], axis=1)
    big_labels = np.zeros(len(path))
    small_labels = np.zeros(len(path))
    # Calculate the distances from each pad. Add label of the pad if distance <= radius
    for pad in BIG_BOOST_POSITIONS:
        distances = np.sqrt(np.square(path.values - pad[:2]).sum(axis=1, dtype=np.float32))
        big_labels += (pad[2] * (distances <= BIG_BOOST_RADIUS))

    for pad in SMALL_BOOST_POSITIONS:
        distances = np.sqrt(np.square(path.values - pad[:2]).sum(axis=1, dtype=np.float32))
        small_labels += (pad[2] * (distances <= SMALL_BOOST_RADIUS))
    # Add labels and exclude labels with z too high. Didn't calculate this earlier because its a flat height)
    df['pad_in_range'] = 0
    df['pad_in_range'] += small_labels
    df.loc[df['pos_z'] >= SMALL_BOOST_HEIGHT, 'pad_in_range'] = 0
    df['pad_in_range'] += big_labels
    df.loc[df['pos_z'] >= BIG_BOOST_HEIGHT, 'pad_in_range'] = 0
    # Get the gains in boost per frame
    df['gains'] = df['boost'].diff().clip(0)
    # Get whether we entered or exited the range of a pad per frame
    df['status_change'] = (df['pad_in_range'].diff(1))
    df = df.fillna(0)
    # Get the index of the frame we most recently entered a pad range, per frame.
    df['recent_entry_index'] = df.index
    df.loc[df['status_change'] <= 0, 'recent_entry_index'] = 0
    df['recent_entry_index'] = df['recent_entry_index'].replace(0, np.nan).fillna(
        method='bfill', limit=LAG_BACK).fillna(
        method='ffill', limit=LAG_FORWARD)
    gains_frames = df.loc[
        ((df['gains'] > 5) & (df['boost'] != 33.33333)) | ((df['gains'] > 0) & (df['boost'] > 95.0))].copy()

    gains_indexes = gains_frames['recent_entry_index'].dropna()

    pickups = df.loc[gains_indexes]['status_change'].copy()
    pickups = pickups.loc[~pickups.index.duplicated(keep='first')]
    return pickups
